import base64
import datetime
import logging
import xml.etree.ElementTree as ET  # noqa: N817
from datetime import timedelta
from pathlib import Path
from typing import Dict, Union

import esdl
from esdl.esdl_handler import EnergySystemHandler

import numpy as np

import pandas as pd

from pyecore.resources import ResourceSet

import rtctools.data.pi as pi
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.io_mixin import IOMixin

from rtctools_heat_network.esdl.asset_to_component_base import _AssetToComponentBase
from rtctools_heat_network.esdl.edr_pipe_class import EDRPipeClass
from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.influxdb.profile import parse_esdl_profiles
from rtctools_heat_network.modelica_component_type_mixin import (
    ModelicaComponentTypeMixin,
)
from rtctools_heat_network.pipe_class import PipeClass
from rtctools_heat_network.pycml.pycml_mixin import PyCMLMixin
from rtctools_heat_network.qth_mixin import QTHMixin

from .common import Asset
from .esdl_heat_model import ESDLHeatModel
from .esdl_qth_model import ESDLQTHModel

logger = logging.getLogger("rtctools_heat_network")


ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}
DEFAULT_START_TIMESTAMP = "2017-01-01T00:00:00+00:00"
DEFAULT_END_TIMESTAMP = "2018-01-01T00:00:00+00:00"


class _ESDLInputException(Exception):
    pass


class ESDLMixin(
    ModelicaComponentTypeMixin,
    IOMixin,
    PyCMLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    esdl_run_info_path: Path = None

    esdl_pi_validate_timeseries = False

    esdl_pi_input_data_config = None
    esdl_pi_output_data_config = None

    __max_supply_temperature = None

    # TODO: remove this once ESDL allows specifying a minimum pipe size for an optional pipe.
    __minimum_pipe_size_name = "DN150"

    def __init__(self, *args, **kwargs):
        esdl_string = kwargs.get("esdl_string", None)
        if esdl_string is not None:
            esdl_path = None
            self.esdl_string = base64.b64decode(esdl_string).decode("utf-8")
            self.__run_info = None
        else:
            self.esdl_string = None
            self.esdl_run_info_path = kwargs.get(
                "esdl_run_info_path", Path(kwargs["input_folder"]) / "RunInfo.xml"
            )

            if not self.esdl_pi_input_data_config:
                self.esdl_pi_input_data_config = _ESDLInputDataConfig

            if not self.esdl_pi_output_data_config:
                self.esdl_pi_output_data_config = _ESDLOutputDataConfig

            self.__run_info = _RunInfoReader(self.esdl_run_info_path)
            esdl_path = self.__run_info.esdl_file

        self.__esdl_assets, self._profiles, self.__esdl_carriers = _esdl_to_assets(
            self.esdl_string, esdl_path
        )

        if self.__run_info is not None and self.__run_info.parameters_file is not None:
            self.__esdl_assets = _overwrite_parameters(
                self.__run_info.parameters_file, self.__esdl_assets
            )

        # This way we allow users to adjust the parsed ESDL assets
        assets = self.esdl_assets

        # Although we work with the names, the FEWS import data uses the component IDs
        self.__timeseries_id_map = {a.id: a.name for a in assets.values()}

        if isinstance(self, HeatMixin):
            self.__model = ESDLHeatModel(assets, **self.esdl_heat_model_options())
        else:
            assert isinstance(self, QTHMixin)

            # Maximum supply temperature is very network dependent, so it is
            # hard to choose a default. Therefore, we look at the global
            # properties instead and add 10 degrees on top.
            global_supply_temperatures = [
                c["supplyTemperature"]
                for a in assets.values()
                for c in a.global_properties["carriers"].values()
            ]
            max_global_supply = max(x for x in global_supply_temperatures if np.isfinite(x))

            attribute_temperatures = [
                a.attributes.get("maxTemperature", -np.inf) for a in assets.values()
            ]
            max_attribute = max(x for x in attribute_temperatures if np.isfinite(x))

            self.__max_supply_temperature = max(max_global_supply, max_attribute) + 10.0

            self.__model = ESDLQTHModel(assets, **self.esdl_qth_model_options())

        root_logger = logging.getLogger("")

        if self.__run_info is not None and self.__run_info.output_diagnostic_file:
            # Add stream handler if it does not already exist.
            if not logger.hasHandlers() and not any(
                (isinstance(h, logging.StreamHandler) for h in logger.handlers)
            ):
                handler = logging.StreamHandler()
                formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
                handler.setFormatter(formatter)
                logger.addHandler(handler)

            # Add pi.DiagHandler. Only add if it doesn't already exist.
            if not any((isinstance(h, pi.DiagHandler) for h in root_logger.handlers)):
                basename = self.__run_info.output_diagnostic_file.stem
                folder = self.__run_info.output_diagnostic_file.parent

                handler = pi.DiagHandler(folder, basename)
                root_logger.addHandler(handler)

        if self.__run_info is not None:
            self._input_timeseries_file = self.__run_info.input_timeseries_file
            self.__output_timeseries_file = self.__run_info.output_timeseries_file

        self._override_pipe_classes = dict()
        self.override_pipe_classes()

        self.name_to_esdl_id_map = dict()

        super().__init__(*args, **kwargs)

    def pre(self):
        super().pre()
        for esdl_id, esdl_asset in self.esdl_assets.items():
            if esdl_asset.name in self.name_to_esdl_id_map:
                raise RuntimeWarning(
                    f"Found multiple ESDL assets with name {esdl_asset.name} in the "
                    f"input ESDL. This is not supported in the optimization."
                )
            self.name_to_esdl_id_map[esdl_asset.name] = esdl_id

    def override_pipe_classes(self):
        maximum_velocity = self.heat_network_options()["maximum_velocity"]

        no_pipe_class = PipeClass("None", 0.0, 0.0, (0.0, 0.0), 0.0)
        pipe_classes = [
            EDRPipeClass.from_edr_class(name, edr_class_name, maximum_velocity)
            for name, edr_class_name in _AssetToComponentBase.STEEL_S1_PIPE_EDR_ASSETS.items()
        ]

        # We assert the pipe classes are monotonically increasing in size
        assert np.all(np.diff([pc.inner_diameter for pc in pipe_classes]) > 0)

        for asset in self.esdl_assets.values():
            if asset.asset_type == "Pipe":
                p = asset.name
                if self.is_cold_pipe(p):
                    continue

                if asset.attributes["state"].name == "OPTIONAL":
                    c = self._override_pipe_classes[p] = []
                    c.append(no_pipe_class)

                    min_size = self.__minimum_pipe_size_name
                    min_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == min_size
                    ]
                    assert len(min_size_idx) == 1
                    min_size_idx = min_size_idx[0]

                    max_size = asset.attributes["diameter"].name

                    max_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == max_size
                    ]
                    assert len(max_size_idx) == 1
                    max_size_idx = max_size_idx[0]

                    if max_size_idx < min_size_idx:
                        logger.warning(
                            f"{p} has an upper DN size smaller than the used minimum size "
                            f"of {self.__minimum_pipe_size_name}, choose at least "
                            f"{self.__minimum_pipe_size_name}"
                        )
                    elif min_size_idx == max_size_idx:
                        c.append(pipe_classes[min_size_idx])
                    else:
                        c.extend(pipe_classes[min_size_idx : max_size_idx + 1])
                elif asset.attributes["state"].name == "DISABLED":
                    c = self._override_pipe_classes[p] = []
                    c.append(no_pipe_class)

    @property
    def esdl_assets(self):
        return self.__esdl_assets

    @property
    def esdl_carriers(self):
        return self.__esdl_carriers

    @property
    def esdl_asset_id_to_name_map(self):
        return self.__timeseries_id_map.copy()

    @property
    def esdl_asset_name_to_id_map(self):
        return dict(zip(self.__timeseries_id_map.values(), self.__timeseries_id_map.keys()))

    def get_asset_from_asset_name(self, asset_name: str) -> esdl.Asset:
        asset_id = self.esdl_asset_name_to_id_map[asset_name]
        return self.esdl_assets[asset_id]

    def esdl_heat_model_options(self) -> Dict:
        heat_network_options = self.heat_network_options()
        v_nominal = heat_network_options["estimated_velocity"]
        v_max = heat_network_options["maximum_velocity"]
        return dict(v_nominal=v_nominal, v_max=v_max)

    def esdl_qth_model_options(self) -> Dict:
        heat_network_options = self.heat_network_options()
        kwargs = {}
        kwargs["v_nominal"] = heat_network_options["estimated_velocity"]
        kwargs["v_max"] = heat_network_options["maximum_velocity"]
        if self.__max_supply_temperature is not None:
            kwargs["maximum_temperature"] = self.__max_supply_temperature
        return dict(**kwargs)

    def is_hot_pipe(self, pipe: str) -> bool:
        return not self.is_cold_pipe(pipe)

    def is_cold_pipe(self, pipe: str) -> bool:
        return pipe.endswith("_ret")

    def hot_to_cold_pipe(self, pipe: str):
        return f"{pipe}_ret"

    def cold_to_hot_pipe(self, pipe: str):
        return pipe[:-4]

    def pycml_model(self):
        return self.__model

    def read(self):
        super().read()

        if self._profiles:
            datetimes = None
            for id, profile in self._profiles.items():
                asset_name = self.esdl_asset_id_to_name_map[id]
                asset = next(a for a in self.esdl_assets.values() if a.name == asset_name)
                ports = []
                ports.extend(asset.in_ports) if asset.in_ports is not None else ports
                ports.extend(asset.out_ports) if asset.out_ports is not None else ports
                if isinstance(ports[0].carrier, esdl.HeatCommodity):
                    commodity = "heat"
                elif isinstance(ports[0].carrier, esdl.ElectricityCommodity):
                    commodity = "electricity"
                elif isinstance(ports[0].carrier, esdl.GasCommodity):
                    commodity = "gas"
                variable = f"{asset_name}.target_{commodity}_demand"
                values = profile.values
                flat_list = []
                for sublist in values:
                    for item in sublist:
                        if not np.isnan(item):
                            flat_list.append(item)
                        else:
                            if len(flat_list) > 0:
                                logger.warning(
                                    f"Found NaN value in profile for {variable},  "
                                    f"using value of previous timestep"
                                )
                                flat_list.append(flat_list[-1])
                            else:
                                logger.error(
                                    f"Found NaN value as first value in profile for "
                                    f"{variable}, using 0 instead"
                                )
                                flat_list.append(0.0)
                if datetimes is None:
                    datetimes = profile.index.tz_convert(None).to_pydatetime()
                elif len(profile.index.tz_convert(None).to_pydatetime()) != len(datetimes):
                    logger.error(f"Unequal profile lengths for {variable}")
                for ensemble_member in range(self.ensemble_size):
                    self.io.set_timeseries(
                        variable, datetimes, flat_list, ensemble_member=ensemble_member
                    )
                if self.io.reference_datetime is None:
                    self.io.reference_datetime = (profile.index.tz_convert(None).to_pydatetime())[0]
        elif not hasattr(self, "_input_timeseries_file"):
            demand_assets = [
                asset for asset in self.esdl_assets.values() if asset.asset_type == "HeatingDemand"
            ]
            try:
                datetimes = self.io.datetimes
            except AttributeError:
                logger.warning(
                    f"No profiles provided for demands, could not infer the period over "
                    f"which to optimize, using the period between "
                    f"{DEFAULT_START_TIMESTAMP} and {DEFAULT_END_TIMESTAMP}."
                )
                start_time = datetime.datetime.fromisoformat(DEFAULT_START_TIMESTAMP)
                end_time = datetime.datetime.fromisoformat(DEFAULT_END_TIMESTAMP)
                datetimes = pd.date_range(start=start_time, end=end_time, freq="H").to_pydatetime()
                if self.io.reference_datetime is None:
                    self.io.reference_datetime = start_time
            for demand in demand_assets:
                if self._profiles and self.esdl_asset_name_to_id_map[demand.name] in self._profiles:
                    continue
                logger.warning(
                    f"No demand profile specified for {demand} in an influxdb "
                    f"profile and no file provided with profiles. Using the asset "
                    f"power instead to generate a constant profile, with default "
                    f"asset power of 0.0"
                )
                demand_power = demand.attributes["power"]
                profile = [demand_power] * len(datetimes)
                for ensemble_member in range(self.ensemble_size):
                    self.io.set_timeseries(
                        variable=f"{demand.name}.target_heat_demand",
                        datetimes=datetimes,
                        values=profile,
                        ensemble_member=ensemble_member,
                        check_duplicates=True,
                    )
        else:
            if self._input_timeseries_file is None:
                return

            input_timeseries_file = Path(self._input_timeseries_file)
            assert input_timeseries_file.is_absolute()
            assert input_timeseries_file.suffix == ".xml" or input_timeseries_file.suffix == ".csv"

            if input_timeseries_file.suffix == ".xml":
                self.read_xml(input_timeseries_file)
            elif input_timeseries_file.suffix == ".csv":
                self.read_csv(input_timeseries_file)

    def read_csv(self, input_timeseries_file):
        csv_data = pd.read_csv(input_timeseries_file)
        try:
            timeseries_import_times = [
                datetime.datetime.strptime(entry.replace("Z", ""), "%Y-%m-%d %H:%M:%S")
                for entry in csv_data["DateTime"].to_numpy()
            ]
        except ValueError:
            try:
                timeseries_import_times = [
                    datetime.datetime.strptime(entry.replace("Z", ""), "%Y-%m-%dT%H:%M:%S")
                    for entry in csv_data["DateTime"].to_numpy()
                ]
            except ValueError:
                try:
                    timeseries_import_times = [
                        datetime.datetime.strptime(entry.replace("Z", ""), "%d-%m-%Y %H:%M")
                        for entry in csv_data["DateTime"].to_numpy()
                    ]
                except ValueError:
                    logger.error("Date time string is not in supported format")

        self.io.reference_datetime = timeseries_import_times[0]
        for ensemble_member in range(self.ensemble_size):
            for demand in self.heat_network_components.get("demand", []):
                try:
                    values = csv_data[f"{demand}"].to_numpy()
                    self.io.set_timeseries(
                        demand + ".target_heat_demand",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass
            for source in self.heat_network_components.get("source", []):
                try:
                    values = csv_data[f"{source.replace(' ', '')}"].to_numpy()
                    self.io.set_timeseries(
                        source + ".target_heat_source",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass
            for demand in self.heat_network_components.get("electricity_demand", []):
                try:
                    values = csv_data[
                        f"{demand.replace(' ', '')}.target_electricity_demand"
                    ].to_numpy()
                    self.io.set_timeseries(
                        demand + ".target_electricity_demand",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass
            for source in self.heat_network_components.get("electricity_source", []):
                try:
                    values = csv_data[
                        f"{source.replace(' ', '')}.target_electricity_source"
                    ].to_numpy()
                    self.io.set_timeseries(
                        source + ".target_electricity_source",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass
            for demand in self.heat_network_components.get("gas_demand", []):
                try:
                    values = csv_data[f"{demand.replace(' ', '')}.target_gas_demand"].to_numpy()
                    self.io.set_timeseries(
                        demand + ".target_gas_demand",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass
            for source in self.heat_network_components.get("gas_source", []):
                try:
                    values = csv_data[f"{source.replace(' ', '')}.target_gas_source"].to_numpy()
                    self.io.set_timeseries(
                        source + ".target_gas_source",
                        timeseries_import_times,
                        values,
                        ensemble_member,
                    )
                except KeyError:
                    pass

    def read_xml(self, input_timeseries_file):
        timeseries_import_basename = input_timeseries_file.stem
        input_folder = input_timeseries_file.parent

        try:
            self.__timeseries_import = pi.Timeseries(
                self.esdl_pi_input_data_config(
                    self.__timeseries_id_map, self.heat_network_components.copy()
                ),
                input_folder,
                timeseries_import_basename,
                binary=False,
                pi_validate_times=self.esdl_pi_validate_timeseries,
            )
        except IOError:
            raise Exception(
                "ESDLMixin: {}.xml not found in {}.".format(
                    timeseries_import_basename, input_folder
                )
            )

        # Convert timeseries timestamps to seconds since t0 for internal use
        timeseries_import_times = self.__timeseries_import.times

        # Offer input timeseries to IOMixin
        self.io.reference_datetime = self.__timeseries_import.forecast_datetime

        for ensemble_member in range(self.__timeseries_import.ensemble_size):
            for variable, values in self.__timeseries_import.items(ensemble_member):
                self.io.set_timeseries(variable, timeseries_import_times, values, ensemble_member)

    def write(self):
        super().write()

        if getattr(self, "__output_timeseries_file", None) is None:
            return

        output_timeseries_file = Path(self.__output_timeseries_file)
        assert output_timeseries_file.is_absolute()
        assert output_timeseries_file.suffix == ".xml"

        timeseries_export_basename = output_timeseries_file.stem
        output_folder = output_timeseries_file.parent

        try:
            self.__timeseries_export = pi.Timeseries(
                self.esdl_pi_output_data_config(self.__timeseries_id_map),
                output_folder,
                timeseries_export_basename,
                binary=False,
                pi_validate_times=self.esdl_pi_validate_timeseries,
            )
        except IOError:
            raise Exception(
                "ESDLMixin: {}.xml not found in {}.".format(
                    timeseries_export_basename, output_folder
                )
            )

        # Get time stamps
        times = self.times()
        if len(set(times[1:] - times[:-1])) == 1:
            dt = timedelta(seconds=times[1] - times[0])
        else:
            dt = None

        output_keys = [k for k, v in self.__timeseries_export.items()]

        # Start of write output
        # Write the time range for the export file.
        self.__timeseries_export.times = [
            self.__timeseries_import.times[self.__timeseries_import.forecast_index]
            + timedelta(seconds=s)
            for s in times
        ]

        # Write other time settings
        self.__timeseries_export.forecast_datetime = self.__timeseries_import.forecast_datetime
        self.__timeseries_export.dt = dt
        self.__timeseries_export.timezone = self.__timeseries_import.timezone

        # Write the ensemble properties for the export file.
        if self.ensemble_size > 1:
            self.__timeseries_export.contains_ensemble = True
        self.__timeseries_export.ensemble_size = self.ensemble_size
        self.__timeseries_export.contains_ensemble = self.ensemble_size > 1

        # Start looping over the ensembles for extraction of the output values.
        for ensemble_member in range(self.ensemble_size):
            results = self.extract_results(ensemble_member)

            # For all variables that are output variables the values are
            # extracted from the results.
            for variable in output_keys:
                try:
                    values = results[variable]
                    if len(values) != len(times):
                        values = self.interpolate(
                            times,
                            self.times(variable),
                            values,
                            self.interpolation_method(variable),
                        )
                except KeyError:
                    try:
                        ts = self.get_timeseries(variable, ensemble_member)
                        if len(ts.times) != len(times):
                            values = self.interpolate(times, ts.times, ts.values)
                        else:
                            values = ts.values
                    except KeyError:
                        logger.warning(
                            "ESDLMixin: Output requested for non-existent variable {}. "
                            "Will not be in output file.".format(variable)
                        )
                        continue

                self.__timeseries_export.set(variable, values, ensemble_member=ensemble_member)

        # Write output file to disk
        self.__timeseries_export.write()


class _ESDLInputDataConfig:
    def __init__(self, id_map, heat_network_components):
        # TODO: change naming source and demand to heat_source and heat_demand throughout code
        self.__id_map = id_map
        self._sources = set(heat_network_components.get("source", []))
        self._demands = set(heat_network_components.get("demand", []))
        self._electricity_sources = set(heat_network_components.get("electricity_source", []))
        self._electricity_demands = set(heat_network_components.get("electricity_demand", []))
        self._gas_sources = set(heat_network_components.get("gas_source", []))
        self._gas_demands = set(heat_network_components.get("gas_demand", []))

    def variable(self, pi_header):
        location_id = pi_header.find("pi:locationId", ns).text

        try:
            component_name = self.__id_map[location_id]
        except KeyError:
            parameter_id = pi_header.find("pi:parameterId", ns).text
            qualifiers = pi_header.findall("pi:qualifierId", ns)
            qualifier_ids = ":".join(q.text for q in qualifiers)
            return f"{location_id}:{parameter_id}:{qualifier_ids}"

        if component_name in self._demands:
            suffix = ".target_heat_demand"
        elif component_name in self._sources:
            suffix = ".target_heat_source"
        elif component_name in self._electricity_demands:
            suffix = ".target_electricity_demand"
        elif component_name in self._electricity_sources:
            suffix = ".target_electricity_source"
        elif component_name in self._gas_demands:
            suffix = ".target_gas_demand"
        elif component_name in self._gas_sources:
            suffix = ".target_gas_source"
        else:
            logger.warning(
                f"Could not identify '{component_name}' as either source or demand. "
                f"Using neutral suffix '.target_heat' for its heat timeseries."
            )
            suffix = ".target_heat"

        # Note that the qualifier id (if any specified) refers to the profile
        # element of the respective ESDL asset->in_port. For now we just
        # assume that only heat demand timeseries are set in the XML file.
        return f"{component_name}{suffix}"

    def pi_variable_ids(self, variable):
        raise NotImplementedError

    def parameter(self, parameter_id, location_id=None, model_id=None):
        raise NotImplementedError

    def pi_parameter_ids(self, parameter):
        raise NotImplementedError


class _ESDLOutputDataConfig:
    def __init__(self, id_map):
        self.__id_map = id_map

    def variable(self, pi_header):
        location_id = pi_header.find("pi:locationId", ns).text
        parameter_id = pi_header.find("pi:parameterId", ns).text

        component_name = self.__id_map[location_id]

        return f"{component_name}.{parameter_id}"

    def pi_variable_ids(self, variable):
        raise NotImplementedError

    def parameter(self, parameter_id, location_id=None, model_id=None):
        raise NotImplementedError

    def pi_parameter_ids(self, parameter):
        raise NotImplementedError


class _RunInfoReader:
    def __init__(self, filepath: Union[str, Path]):
        filepath = Path(filepath).resolve()

        root = ET.parse(filepath).getroot()

        # If the workDir is not absolute, we take it relative to the folder in
        # which the RunInfo file is.
        work_dir = Path(root.findtext("pi:workDir", namespaces=ns))
        if not work_dir.is_absolute():
            work_dir = filepath.parent / work_dir

        self.esdl_file = Path(root.find("pi:properties", ns)[0].attrib["value"])
        if not self.esdl_file.is_absolute():
            self.esdl_file = work_dir / self.esdl_file

        self.parameters_file = root.findtext("pi:inputParameterFile", namespaces=ns)
        if self.parameters_file is not None:
            self.parameters_file = Path(self.parameters_file)
            if not self.parameters_file.is_absolute():
                self.parameters_file = work_dir / self.parameters_file

        try:
            self.input_timeseries_file = Path(root.findall("pi:inputTimeSeriesFile", ns)[0].text)
            if not self.input_timeseries_file.is_absolute():
                self.input_timeseries_file = work_dir / self.input_timeseries_file
        except IndexError:
            self.input_timeseries_file = None

        try:
            self.output_timeseries_file = Path(root.findall("pi:outputTimeSeriesFile", ns)[0].text)
            if not self.output_timeseries_file.is_absolute():
                self.output_timeseries_file = work_dir / self.output_timeseries_file
        except IndexError:
            self.output_timeseries_file = None

        try:
            self.output_diagnostic_file = Path(root.findall("pi:outputDiagnosticFile", ns)[0].text)
            if not self.output_diagnostic_file.is_absolute():
                self.output_diagnostic_file = work_dir / self.output_diagnostic_file
        except IndexError:
            self.output_diagnostic_file = None


def _overwrite_parameters(parameters_file, assets):
    paramroot = ET.parse(parameters_file).getroot()
    groups = paramroot.findall("pi:group", ns)

    for parameter in groups:
        id_ = parameter.attrib["id"]
        param_name = parameter[0].attrib["id"]
        param_value = parameter[0][0].text

        asset = assets[id_]
        type_ = type(asset.attributes[param_name])
        asset.attributes[param_name] = type_(param_value)

    return assets


def _esdl_to_assets(esdl_string, esdl_path: Union[Path, str]):
    if esdl_string is None:
        # correct profile attribute
        esdl.ProfileElement.from_.name = "from"
        setattr(esdl.ProfileElement, "from", esdl.ProfileElement.from_)

        # using esdl as resourceset
        rset_existing = ResourceSet()

        # read esdl energy system
        resource_existing = rset_existing.get_resource(str(esdl_path))
        created_energy_system = resource_existing.contents[0]

        esdl_model = created_energy_system
    else:
        esh = EnergySystemHandler()
        esdl_model = esh.load_from_string(esdl_string)

    # global properties
    global_properties = {}
    global_properties["carriers"] = {}
    id_to_idnumber_map = {}

    for x in esdl_model.energySystemInformation.carriers.carrier.items:
        if isinstance(x, esdl.esdl.HeatCommodity):
            if x.supplyTemperature != 0.0 and x.returnTemperature == 0.0:
                type_ = "supply"
            elif x.returnTemperature != 0.0 and x.supplyTemperature == 0.0:
                type_ = "return"
            else:
                type_ = "none"
            if x.id not in id_to_idnumber_map:
                number_list = [int(s) for s in x.id if s.isdigit()]
                number = ""
                for nr in number_list:
                    number = number + str(nr)
                if type_ == "return":
                    number = number + "000"
                id_to_idnumber_map[x.id] = int(number)

            global_properties["carriers"][x.id] = dict(
                name=x.name.replace("_ret", ""),
                id=x.id,
                id_number_mapping=id_to_idnumber_map[x.id],
                supplyTemperature=x.supplyTemperature,
                returnTemperature=x.returnTemperature,
                __rtc_type=type_,
            )

    # For now, we only support networks with two carries; one hot, one cold.
    # When this no longer holds, carriers either have to specify both the
    # supply and return temperature (instead of one being 0.0), or we have to
    # pair them up.
    if (len(global_properties["carriers"]) % 2) != 0:
        _ESDLInputException(
            "Odd number of carriers specified, please use model with dedicated supply and return "
            "carriers. Every hydraulically coupled system should have one carrier for the supply "
            "side and one for the return side"
        )

    for c in global_properties["carriers"].values():
        supply_temperature = next(
            x["supplyTemperature"]
            for x in global_properties["carriers"].values()
            if x["supplyTemperature"] != 0.0 and x["name"] == c["name"]
        )
        return_temperature = next(
            x["returnTemperature"]
            for x in global_properties["carriers"].values()
            if x["returnTemperature"] != 0.0 and x["name"] == c["name"]
        )
        c["supplyTemperature"] = supply_temperature
        c["returnTemperature"] = return_temperature

    for x in esdl_model.energySystemInformation.carriers.carrier.items:
        if isinstance(x, esdl.esdl.ElectricityCommodity):
            global_properties["carriers"][x.id] = dict(
                name=x.name,
                voltage=x.voltage,
            )

    assets = {}

    # Component ids are unique, but we require component names to be unique as well.
    component_names = set()

    # loop through assets
    for el in esdl_model.eAllContents():
        if isinstance(el, esdl.Asset):
            if hasattr(el, "name") and el.name:
                el_name = el.name
            else:
                el_name = el.id

            if "." in el_name:
                # Dots indicate hierarchy, so would be very confusing
                raise ValueError(f"Dots in component names not supported: '{el_name}'")

            if el_name in component_names:
                raise Exception(f"Asset names have to be unique: '{el_name}' already exists")
            else:
                component_names.add(el_name)

            # For some reason `esdl_element.assetType` is `None`, so use the class name
            asset_type = el.__class__.__name__

            # Every asset should at least have a port to be connected to another asset
            assert len(el.port) >= 1

            in_ports = None
            out_ports = None
            for port in el.port:
                if isinstance(port, esdl.InPort):
                    if in_ports is None:
                        in_ports = [port]
                    else:
                        in_ports.append(port)
                elif isinstance(port, esdl.OutPort):
                    if out_ports is None:
                        out_ports = [port]
                    else:
                        out_ports.append(port)
                else:
                    _ESDLInputException(f"The port for {el_name} is neither an IN or OUT port")

            # Note that e.g. el.__dict__['length'] does not work to get the length of a pipe.
            # We therefore built this dict ourselves using 'dir' and 'getattr'
            attributes = {k: getattr(el, k) for k in dir(el)}
            assets[el.id] = Asset(
                asset_type,
                el.id,
                el_name,
                in_ports,
                out_ports,
                attributes,
                global_properties,
            )
    profiles = parse_esdl_profiles(esdl_model)

    return assets, profiles, global_properties["carriers"]



from rtctools_heat_network.pycml import Variable

from .gas_base import GasPort
from .._internal import BaseAsset
from .._internal.gas_component import GasComponent


class GasNode(GasComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_node"

        self.n = 2
        assert self.n >= 2

        self.add_variable(GasPort, "GasConn", self.n)
        self.add_variable(Variable, "H", min=0.0)

        for i in range(1, self.n + 1):
            self.add_equation(self.GasConn[i].H - self.H)

        # Because the orientation of the connected pipes are important to setup the mass
        # conservation, these constraints are added in the mixin.



from numpy import nan


from rtctools_heat_network.pycml import Variable

from ._internal.heat_component import BaseAsset
from .heat_two_port import HeatTwoPort


class ATES(HeatTwoPort, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "ates"

        self.Q_nominal = 1.0
        self.T_supply = nan
        self.T_return = nan
        self.T_supply_id = -1
        self.T_return_id = -1
        self.dT = self.T_supply - self.T_return
        self.cp = 4200.0
        self.rho = 988.0
        self.Heat_nominal = self.cp * self.rho * self.dT * self.Q_nominal

        self.heat_loss_coeff = 0.005 / (24.0 * 3600.0)
        self.single_doublet_power = nan
        self.nr_of_doublets = 1.0
        # The hot/cold tank can have a lower bound on its volume.
        # Meaning that they might always be, for e.g., 5% full.
        self.min_fraction_tank_volume = 0.05

        # Stored_heat is the heat that is contained in the ates.
        # Heat_ates is the amount of heat added to or extracted from the buffer
        # per timestep.
        # Thus Heat_buffer = HeatHot = der(Stored_heat).
        # We connect an ATES as an demand, meaning that flow and Heat_ates are positive under
        # charging and negative under discharge
        self.add_variable(Variable, "Heat_ates", nominal=self.Heat_nominal)
        self.add_variable(Variable, "Heat_flow", nominal=self.Heat_nominal)
        # Assume the storage fills in about 3 months at typical rate
        self._typical_fill_time = 3600.0 * 24.0 * 90.0
        self._nominal_stored_heat = self.Heat_nominal * self._typical_fill_time
        self.add_variable(
            Variable,
            "Stored_heat",
            min=0.0,
            nominal=self._nominal_stored_heat,
        )
        self.add_variable(Variable, "Q", nominal=self.Q_nominal)
        # For nicer constraint coefficient scaling, we shift a bit more error into
        # the state vector entry of `Heat_loss`. In other words, with a factor of
        # 10.0, we aim for a state vector entry of ~0.1 (instead of 1.0)
        self._heat_loss_error_to_state_factor = 10.0
        self._nominal_heat_loss = (
            self._nominal_stored_heat * self.heat_loss_coeff * self._heat_loss_error_to_state_factor
        )
        self.add_variable(Variable, "Heat_loss", min=0.0, nominal=self._nominal_heat_loss)

        self._heat_loss_eq_nominal_ates = (self.Heat_nominal * self._nominal_heat_loss) ** 0.5

        self.add_equation(self.HeatIn.Q - self.HeatOut.Q)
        self.add_equation(self.Q - self.HeatOut.Q)

        # Heat stored in the ates
        self.add_equation(
            (self.der(self.Stored_heat) - self.Heat_ates + self.Heat_loss)
            / self._heat_loss_eq_nominal_ates
        )
        self.add_equation(
            (self.Heat_loss - self.Stored_heat * self.heat_loss_coeff) / self._nominal_heat_loss
        )
        self.add_equation(
            (self.HeatIn.Heat - (self.HeatOut.Heat + self.Heat_ates)) / self.Heat_nominal
        )
        self.add_equation((self.Heat_flow - self.Heat_ates) / self.Heat_nominal)



from rtctools_heat_network.pycml import Connector, Variable

from .._internal.electricity_component import ElectricityComponent


class ElectricityPort(ElectricityComponent, Connector):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "Power")
        self.add_variable(Variable, "V", min=0.0)
        self.add_variable(Variable, "I")


class ElectricityTwoPort(ElectricityComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(ElectricityPort, "ElectricityIn")
        self.add_variable(ElectricityPort, "ElectricityOut")



from numpy import nan, pi

from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class Pipe(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "pipe"
        self.disconnectable = False
        self.has_control_valve = False

        self.length = 1.0
        self.diameter = 1.0
        assert "area" not in modifiers, "modifying area directly is not allowed"
        self.area = 0.25 * pi * self.diameter**2
        self.temperature = nan

        # Parameters determining the heat loss
        # All of these have default values in the library function
        self.insulation_thickness = nan
        self.conductivity_insulation = nan
        self.conductivity_subsoil = nan
        self.depth = nan
        self.h_surface = nan
        self.pipe_pair_distance = nan

        self.T_ground = 10.0

        self.Heat_loss = nan

        self.add_variable(Variable, "dH")

        # rho * ff * length * area / 2 / diameter * velocity**3
        ff = 0.02  # Order of magnitude expected with 0.05-2.5m/s in 20mm-1200mm diameter pipe
        velo = self.Q_nominal / self.area
        self.Hydraulic_power_nominal = (
            self.rho * ff * self.length * pi * self.area / self.diameter / 2.0 * velo**3
        )
        self.add_variable(
            Variable, "Hydraulic_power", min=0.0, nominal=self.Hydraulic_power_nominal
        )  # [W]

        self.add_equation(((self.Heat_flow - self.HeatIn.Heat) / self.Heat_nominal))

        # Note: Heat loss is added in the mixin, because it depends on the flow direction
        # * heat loss equation: (HeatOut.Heat - (HeatIn.Heat +/- Heat_loss)) = 0.



from rtctools_heat_network.pycml import Variable

from .electricity_base import ElectricityTwoPort
from .._internal import BaseAsset


class ElectricityCable(ElectricityTwoPort, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_cable"
        self.disconnectable = False

        self.add_variable(Variable, "Power_loss")

        self.length = 1.0

        # Powerloss with inequality in the heat-mixin
        # values for NAYY 4x50 SE
        # from: https://pandapower.readthedocs.io/en/v2.6.0/std_types/basic.html
        self.max_current = 142.0
        self.max_voltage = 1000.0
        self.min_voltage = 230.0
        self.nominal_current = self.max_current / 2.0
        self.nominal_voltage = (self.max_voltage + self.min_voltage) / 2.0
        self.r = 1.0e-6 * self.length  # TODO: temporary value

        self.add_equation(
            (
                (self.ElectricityOut.V - (self.ElectricityIn.V - self.r * self.ElectricityIn.I))
                / ((self.nominal_voltage * self.r * self.nominal_current) ** 0.5)
            )
        )
        self.add_equation(((self.ElectricityIn.I - self.ElectricityOut.I) / self.nominal_current))
        self.add_equation(
            (
                (self.ElectricityOut.Power - (self.ElectricityIn.Power - self.Power_loss))
                / (self.nominal_voltage * self.nominal_current)
            )
        )



from rtctools_heat_network.pycml import Variable

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricityNode(ElectricityComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_node"

        self.n = 2
        assert self.n >= 2

        self.add_variable(ElectricityPort, "ElectricityConn", self.n)
        self.add_variable(Variable, "V", min=0.0)

        for i in range(1, self.n + 1):
            self.add_equation(self.ElectricityConn[i].V - self.V)

        # Because the orientation of the connected cables are important to setup the energy
        # conservation, these constraints are added in the mixin.



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestVaryingTemperature(TestCase):
    def test_1a_temperature_variation(self):
        import models.unit_cases.case_1a.src.run_1a as run_1a
        from models.unit_cases.case_1a.src.run_1a import HeatProblemTvar

        base_folder = Path(run_1a.__file__).resolve().parent.parent

        heat_problem = run_optimization_problem(HeatProblemTvar, base_folder=base_folder)

        # This optimization problem is to see whether the correct minimum delta temperaute is
        # chose by the optimization, minimum of 21 deg is needed, only the 75/60 option is
        # feasible, note that a maximum velocity is set low to achieve this.
        results = heat_problem.extract_results()

        # Check that the highest supply temperature is selected
        np.testing.assert_allclose(results[f"{3625334968694477359}__supply_temperature"], 75.0)

        # Check that the lowest return temperature is selected
        np.testing.assert_allclose(results[f"{3625334968694477359000}__return_temperature"], 60.0)

    def test_3a_temperature_variation(self):
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemTvarsup, HeatProblemTvarret

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        heat_problem = run_optimization_problem(HeatProblemTvarsup, base_folder=base_folder)

        # optimization with two choices in supply temp 80 and 120 deg
        # lowest temperature should be selected because of lower heat losses and
        # heat production minimization goal
        results = heat_problem.extract_results()

        # Check whehter the heat demand is matched
        for d in heat_problem.heat_network_components.get("demand", []):
            target = heat_problem.get_timeseries(f"{d}.target_heat_demand").values[
                : len(heat_problem.times())
            ]
            np.testing.assert_allclose(target, results[f"{d}.Heat_demand"])

        # Check that the lowest temperature (80.0) is the outputted temperature
        np.testing.assert_allclose(results[f"{4195016129475469474608}__supply_temperature"], 80.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{4195016129475469474608}__supply_80.0"], 1.0)
        np.testing.assert_allclose(results[f"{4195016129475469474608}__supply_120.0"], 0.0)

        heat_problem = run_optimization_problem(HeatProblemTvarret, base_folder=base_folder)

        # optimization with two choices in return temp 30 and 40 deg
        # lowest temperature should be selected because of larger dT causing lowest flow rates
        # and we apply source Q minimization goal
        results = heat_problem.extract_results()

        # Check whehter the heat demand is matched
        for d in heat_problem.heat_network_components.get("demand", []):
            target = heat_problem.get_timeseries(f"{d}.target_heat_demand").values[
                : len(heat_problem.times())
            ]
            np.testing.assert_allclose(target, results[f"{d}.Heat_demand"])

        # Check that the lowest temperature (30.0) is the outputted temperature
        np.testing.assert_allclose(
            results[f"{4195016129475469474608000}__return_temperature"], 30.0
        )
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{4195016129475469474608000}__return_30.0"], 1.0)
        np.testing.assert_allclose(results[f"{4195016129475469474608000}__return_40.0"], 0.0)

    def test_hex_temperature_variation(self):
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblemTvar,
            HeatProblemTvarDisableHEX,
            HeatProblemTvarSecondary,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        heat_problem = run_optimization_problem(HeatProblemTvar, base_folder=base_folder)

        # optimization with three choices of primary supply temperature of which the lowest is
        # infeasible. Therefore optimization should select second lowest option of 80.
        # lowest feasible temperature should be selected due to heatlosses and the minimization
        # goal of the sources
        results = heat_problem.extract_results()

        # Check that the lowest feasible temperature (80.0) is the outputted temperature and not
        # the 69 which is below the secondary side supply temperature
        np.testing.assert_allclose(results[f"{33638164429859421}__supply_temperature"], 80.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{33638164429859421}__supply_69.0"], 0.0)
        np.testing.assert_allclose(results[f"{33638164429859421}__supply_80.0"], 1.0)
        np.testing.assert_allclose(results[f"{33638164429859421}__supply_90.0"], 0.0)

        heat_problem = run_optimization_problem(HeatProblemTvarDisableHEX, base_folder=base_folder)

        # optimization with only one option in temperature which is infeasible for the hex.
        # therefore optimization should disable the heat exchanger
        results = heat_problem.extract_results()

        # Check that the problem has an infeasible temperature for the hex
        np.testing.assert_allclose(results[f"{33638164429859421}__supply_temperature"], 69.0)
        # Verify that the hex is disabled
        np.testing.assert_allclose(results["HeatExchange_39ed__disabled"], 1.0)
        np.testing.assert_allclose(results["HeatExchange_39ed.Primary_heat"], 0.0)

        heat_problem = run_optimization_problem(HeatProblemTvarSecondary, base_folder=base_folder)

        # optimization with two choices in secondary supply temp 70 and 90 deg
        # lowest temperature should be selected because of larger dT causing lowest flow rates
        # and we apply source dH minimization goal
        results = heat_problem.extract_results()

        # Check that the lowest temperature (70.0) is the outputted temperature
        np.testing.assert_allclose(results[f"{7212673879469902607010}__supply_temperature"], 70.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{7212673879469902607010}__supply_70.0"], 1.0)
        np.testing.assert_allclose(results[f"{7212673879469902607010}__supply_90.0"], 0.0)

    # Note that CBC struggles heavily and tends to crash, therefore excluded from pipeline
    # def test_varying_temperature_with_pipe_sizing(self):
    #     root_folder = str(Path(__file__).resolve().parent.parent)
    #     import sys
    #     sys.path.insert(1, root_folder)
    #
    #     import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
    #     from examples.pipe_diameter_sizing.src.example import (
    #         PipeDiameterSizingProblemTvar,
    #     )  # noqa: E402, I100
    #
    #     base_folder = (
    #         Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
    #     )
    #
    #     del root_folder
    #     sys.path.pop(1)
    #
    #     # The optimization should select the highest delta-temperature as
    #     # this minimizes the length
    #     # times diameter cost function
    #
    #     # problem = run_optimization_problem(
    #     PipeDiameterSizingProblemTvar, base_folder=base_folder)
    #     # results = problem.extract_results()
    #
    #
    #     # # Check that the highest supply temperature is selected
    #     # np.testing.assert_allclose(
    #     results[f"{761602374459208051248}__supply_temperature"], 90.0)
    #     #
    #     # # Check that the lowest return temperature is selected
    #     # np.testing.assert_allclose(
    #     results[f"{761602374459208051248000}__return_temperature"], 30.0)



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.head_loss_mixin import HeadLossOption
from rtctools_heat_network.util import run_heat_network_optimization


class TestHeadLossCalculation(TestCase):
    def test_scalar_return_type(self):
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatPython

        class Model(HeatPython):
            def __init__(self, head_loss_option, *args, **kwargs):
                self.__head_loss_option = head_loss_option
                super().__init__(*args, **kwargs)

            def _hn_get_pipe_head_loss_option(self, *args, **kwargs):
                return self.__head_loss_option

            def optimize(self):
                # Just pre, we don't care about anything else
                self.pre()

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent

        for h in [
            HeadLossOption.LINEAR,
            HeadLossOption.CQ2_INEQUALITY,
            HeadLossOption.LINEARIZED_DW,
        ]:
            m = run_optimization_problem(Model, head_loss_option=h, base_folder=base_folder)

            options = m.heat_network_options()
            parameters = m.parameters(0)

            ret = m._hn_pipe_head_loss("pipe_hot", options, parameters, 0.1)
            self.assertIsInstance(ret, float)

            ret = m._hn_pipe_head_loss("pipe_hot", options, parameters, np.array([0.1]))
            self.assertIsInstance(ret, np.ndarray)
            self.assertEqual(len(ret), 1)

            ret = m._hn_pipe_head_loss("pipe_hot", options, parameters, np.array([0.05, 0.1, 0.2]))
            self.assertIsInstance(ret, np.ndarray)
            self.assertEqual(len(ret), 3)


class TestHeadLossOptions(TestCase):
    def test_no_head_loss_mixing_options(self):
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatPython

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent

        class Model(HeatPython):
            def heat_network_options(self):
                options = super().heat_network_options()
                options["head_loss_option"] = HeadLossOption.LINEAR
                return options

            def _hn_get_pipe_head_loss_option(self, *args, **kwargs):
                return HeadLossOption.NO_HEADLOSS

        with self.assertRaisesRegex(
            Exception, "Mixing .NO_HEADLOSS with other head loss options is not allowed"
        ):
            run_optimization_problem(Model, base_folder=base_folder)

    def test_no_head_loss(self):
        # Test if a model with NO_HEADLOSS set runs without issues
        from models.basic_buffer.src.compare import (
            HeatProblemPyCML,
            QTHProblemPyCML,
            base_folder,
        )

        class ModelHeat(HeatProblemPyCML):
            def heat_network_options(self):
                options = super().heat_network_options()
                options["head_loss_option"] = HeadLossOption.NO_HEADLOSS
                return options

        class ModelQTH(QTHProblemPyCML):
            def heat_network_options(self):
                options = super().heat_network_options()
                options["head_loss_option"] = HeadLossOption.NO_HEADLOSS
                return options

        run_heat_network_optimization(ModelHeat, ModelQTH, base_folder=base_folder)



import logging
from datetime import timedelta as td

import esdl

from influxdb import InfluxDBClient

import pandas as pd

import requests.exceptions


logger = logging.getLogger()


data_set = {}
influx_cred_map = {"wu-profiles.esdl-beta.hesi.energy:443": ("warmingup", "warmingup")}
time_step = td(seconds=3600)
time_step_notation = "{}s".format(int(time_step.total_seconds()))


def get_mw_multiplier(profile: esdl.GenericProfile):
    ts = time_step.total_seconds()
    factor = {
        "ENERGY_IN_WH": 3.6 / (1e3 * ts),
        "ENERGY_IN_KWH": 3.6 / ts,
        "ENERGY_IN_MWH": (3.6 * 1e3) / ts,
        "ENERGY_IN_GWH": (3.6 * 1e6) / ts,
        "ENERGY_IN_TWH": (3.6 * 1e9) / ts,
        "ENERGY_IN_PWH": (3.6 * 1e12) / ts,
        "ENERGY_IN_J": 1 / (1e6 * ts),
        "ENERGY_IN_KJ": 1 / (1e3 * ts),
        "ENERGY_IN_MJ": 1 / ts,
        "ENERGY_IN_GJ": 1e3 / ts,
        "ENERGY_IN_TJ": 1e6 / ts,
        "ENERGY_IN_PJ": 1e9 / ts,
        "POWER_IN_W": 1 / 1e6,
        "POWER_IN_KW": 1 / 1e3,
        "POWER_IN_MW": 1,
        "POWER_IN_GW": 1e3,
        "POWER_IN_TW": 1e6,
        "POWER_IN_PW": 1e9,
    }

    mult_suffix = {
        "NONE": "",
        "KILO": "K",
        "MEGA": "M",
        "GIGA": "G",
        "TERRA": "T",
        "PETA": "P",
        "MILLI": "-",
        "MICRO": "-",
        "NANO": "-",
        "PICO": "-",
    }

    if profile.profileType is not None and profile.profileType != esdl.ProfileTypeEnum.UNDEFINED:
        if profile.profileType.name not in factor:
            print("Unsupported profile type : {}".format(profile.profileType))
            return 0
        return factor[profile.profileType.name]

    if isinstance(profile.profileQuantityAndUnit, esdl.QuantityAndUnitReference):
        p = profile.profileQuantityAndUnit.reference
    elif isinstance(profile.profileQuantityAndUnit, esdl.QuantityAndUnitType):
        p = profile.profileQuantityAndUnit

    if p is not None:
        if p.unit == esdl.UnitEnum.JOULE:
            phy_quantity = "ENERGY_IN_"
            suffix = "J"
        elif p.unit == esdl.UnitEnum.WATTHOUR:
            phy_quantity = "ENERGY_IN_"
            suffix = "WH"
        elif p.unit == esdl.UnitEnum.WATT:
            phy_quantity = "POWER_IN_"
            suffix = "W"
        else:
            print("Unsupported unit : {}".format(p.unit))
            return 0

        prefix = mult_suffix[p.multiplier.name]
        if prefix == "-":
            print("Unsupported multiplier : {}".format(p.multiplier))
            return 0

        return factor["{}{}{}".format(phy_quantity, prefix, suffix)]


def parse_esdl_profiles(es, start_date=None, end_date=None):
    logger.info("Caching profiles...")
    error_neighbourhoods = list()
    for profile in [x for x in es.eAllContents() if isinstance(x, esdl.InfluxDBProfile)]:
        # if profile.eContainer().carrier.id == self.carrier_id:
        profile_host = profile.host
        containing_asset_id = profile.eContainer().energyasset.id
        to_mw_multiplier = get_mw_multiplier(profile)
        # print('Processing profile of {}'.format(containing_asset_id))
        ssl_setting = False
        if "https" in profile_host:
            profile_host = profile_host[8:]
            ssl_setting = True
        elif "http" in profile_host:
            profile_host = profile_host[7:]
        if profile.port == 443:
            ssl_setting = True
        influx_host = "{}:{}".format(profile_host, profile.port)
        if influx_host in influx_cred_map:
            (username, password) = influx_cred_map[influx_host]
        else:
            username = None
            password = None
        client = InfluxDBClient(
            host=profile_host,
            port=profile.port,
            username=username,
            password=password,
            ssl=ssl_setting,
            verify_ssl=ssl_setting,
            timeout=1,
        )
        client.switch_database(profile.database)
        # TODO: check these timebounds how they work.
        # if profile.startDate is not None:
        start_date = profile.startDate.isoformat()
        start_date = start_date if start_date.endswith("+00:00") else start_date + "+00:00"
        # else:
        #     start_date = start_date.isoformat()
        # if profile.endDate is not None:
        end_date_suffix = " AND time <= '{}'".format(profile.endDate.isoformat())
        end_date_suffix = (
            end_date_suffix
            if end_date_suffix.endswith("+00:00'")
            else end_date_suffix[:-1] + "+00:00'"
        )
        # else:
        #     end_date_suffix = ' AND time <= \'{}\''.format(end_date.isoformat())
        if profile.filters is not None and profile.filters != "":
            filter_suffix = " AND {}".format(profile.filters)
        else:
            filter_suffix = ""

        query = 'SELECT MEAN("{}") FROM "{}" WHERE time >= \'{}\'{}{} GROUP BY time({})'.format(
            profile.field,
            profile.measurement,
            start_date,
            end_date_suffix,
            filter_suffix,
            time_step_notation,
        )
        logger.debug("Executing query...\n{}".format(query))
        try:
            data = client.query(query=query)
        except requests.exceptions.ConnectTimeout:
            error_neighbourhoods.append(profile.field)
            logger.error(
                f"Timeout on query for {profile.field}, probably because either the host "
                f"{profile_host} can't be reached or the {profile.port} is wrong."
            )
        data_points = {t[0]: t[1] for t in data.raw["series"][0]["values"]}
        df = pd.DataFrame.from_dict(data_points, orient="index")
        df.index = pd.to_datetime(df.index, utc=True)
        data_set[containing_asset_id] = df * profile.multiplier * to_mw_multiplier * 1.0e6
        client.close()

        if len(error_neighbourhoods) > 0:
            raise RuntimeError(f"Encountered errors loading data for {error_neighbourhoods}")
    return data_set



from pathlib import Path
from unittest import TestCase

import numpy as np


from rtctools.util import run_optimization_problem


class TestMILPGasMultiDemandSourceNode(TestCase):
    """Unit tests for the MILP test case of 2 demands, 2 sources and a node"""

    def test_multi_demand_source_node(self):
        """Test to verify that head is equal for all ports at a node. And throughout the network
        the flow (mass) balance is maintained"""
        import models.unit_cases_gas.multi_demand_source_node.src.run_test as example
        from models.unit_cases_gas.multi_demand_source_node.src.run_test import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        results = run_optimization_problem(GasProblem, base_folder=base_folder).extract_results()

        # Test head at node
        for i in range(1, 5):
            np.testing.assert_allclose(
                results[f"Joint_17c4.GasConn[{i}].H"], results["Joint_17c4.H"]
            )

        # Test if head is going down
        np.testing.assert_allclose(
            results["GasDemand_47d0.Gas_demand_flow"] + results["GasDemand_7978.Gas_demand_flow"],
            results["GasProducer_3573.Gas_source_flow"]
            + results["GasProducer_a977.Gas_source_flow"],
        )



import math

from rtctools_heat_network.pycml import Variable

from .gas_base import GasTwoPort
from .._internal import BaseAsset


class GasPipe(GasTwoPort, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_pipe"
        self.disconnectable = False

        self.nominal_head = 30.0
        self.length = 1.0
        self.nominal_flow_velocity = 1.0
        self.diameter = 1.0
        self.r = 1.0e-6 * self.length  # TODO: temporary value
        self.nominal_flow = (self.diameter / 2) ** 2 * math.pi * self.nominal_flow_velocity
        self.nominal_head_loss = (self.nominal_flow * self.r * self.nominal_head) ** 0.5

        self.add_variable(Variable, "dH", nominal=self.nominal_flow * self.r)

        # head is lost over the pipe
        self.add_equation(((self.GasOut.H - (self.GasIn.H - self.dH)) / self.nominal_head))
        # for now simple linear head loss
        self.add_equation(((self.dH - self.GasIn.Q * self.r) / self.nominal_head_loss))
        # Flow should be preserved
        self.add_equation(((self.GasIn.Q - self.GasOut.Q) / self.nominal_flow))
        # shadow Q for aliases
        self.add_equation(
            ((self.GasOut.Q_shadow - (self.GasIn.Q_shadow - 1.0e-3)) / self.nominal_flow)
        )



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestMultipleCarriers(TestCase):
    def test_multiple_carriers(self):
        import models.multiple_carriers.src.run_multiple_carriers as run_multiple_carriers
        from models.multiple_carriers.src.run_multiple_carriers import (
            HeatProblem,
        )

        base_folder = Path(run_multiple_carriers.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        heat_demand_3222 = results["HeatingDemand_3322.Heat_demand"]
        heat_demand_18aa = results["HeatingDemand_18aa.Heat_demand"]
        heat_demand_3222_q = results["HeatingDemand_3322.Q"]
        heat_demand_18aa_q = results["HeatingDemand_18aa.Q"]

        # Values used in non_storage_component.py
        cp = 4200.0
        rho = 988.0

        # We check for a system consisting out of 2 hydraulically decoupled networks that the energy
        # balance equations are done with the correct carrier.
        np.testing.assert_allclose(heat_demand_3222, heat_demand_3222_q * cp * rho * 30.0)
        np.testing.assert_allclose(heat_demand_18aa, heat_demand_18aa_q * cp * rho * 40.0)



from ._internal import QTHComponent
from .qth_port import QTHPort


class QTHTwoPort(QTHComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(QTHPort, "QTHIn")
        self.add_variable(QTHPort, "QTHOut")



from unittest import TestCase

import numpy as np

from rtctools_heat_network.pycml import Model, Variable


class TestPyCML(TestCase):
    def test_merge_modifiers_function(self):
        x = dict(V=10.0, OutPort=dict(Q=dict(min=0, nominal=1.0)))
        y = dict(OutPort=dict(Q=dict(min=0.5)))

        normal_merge = {**x, **y}
        self.assertEqual(normal_merge, dict(V=10.0, OutPort=dict(Q=dict(min=0.5))))

        recursive_merge = Model.merge_modifiers(x, y)

        self.assertEqual(recursive_merge, dict(V=10.0, OutPort=dict(Q=dict(min=0.5, nominal=1.0))))

    def test_merge_modifiers_models(self):
        class Port(Model):
            def __init__(self, name, **modifiers):
                super().__init__(name, **modifiers)

                self.add_variable(Variable, "Q", min=1.0, nominal=2.0, max=3.0)

        class TwoPort(Model):
            def __init__(self, name, **modifiers):
                super().__init__(name, **modifiers)

                self.add_variable(Port, "InPort", Q=dict(min=1.5))
                self.add_variable(Port, "OutPort", Q=dict(min=0.5, max=2.5))

        class Storage(TwoPort):
            def __init__(self, name, **modifiers):
                super().__init__(name, **modifiers)

                self.add_variable(Variable, "V", min=0.0)

        class A(Model):
            """
            No additional modifiers
            """

            def __init__(self):
                super().__init__(None)

                self.add_variable(Storage, "storage")

        class B(Model):
            """
            Additional modifiers on all variables
            """

            def __init__(self):
                super().__init__(None)

                self.add_variable(
                    Storage, "storage", V=dict(min=10.0), OutPort=dict(Q=dict(nominal=1.5, max=4.0))
                )

        # We testing/comparing with min, max and nominals
        def _var_min_max_nominal(v):
            return dict(min=v.min, nominal=v.nominal, max=v.max)

        a = A()
        a_flat = a.flatten()

        self.assertEqual(
            _var_min_max_nominal(a_flat.variables["storage.InPort.Q"]),
            dict(min=1.5, nominal=2.0, max=3.0),
        )
        self.assertEqual(
            _var_min_max_nominal(a_flat.variables["storage.OutPort.Q"]),
            dict(min=0.5, nominal=2.0, max=2.5),
        )
        self.assertEqual(
            _var_min_max_nominal(a_flat.variables["storage.V"]),
            dict(min=0.0, nominal=0, max=np.inf),
        )

        b = B()
        b_flat = b.flatten()

        self.assertEqual(
            _var_min_max_nominal(b_flat.variables["storage.InPort.Q"]),
            dict(min=1.5, nominal=2.0, max=3.0),
        )
        self.assertEqual(
            _var_min_max_nominal(b_flat.variables["storage.OutPort.Q"]),
            dict(min=0.5, nominal=1.5, max=4.0),
        )
        self.assertEqual(
            _var_min_max_nominal(b_flat.variables["storage.V"]),
            dict(min=10.0, nominal=0, max=np.inf),
        )
